part of '../export.dart';

class RawLens<AIn, AOut, BIn, BOut> {
  final Function1<AIn, BOut> get;
  final Function2<AIn, BIn, AOut> update;

  const RawLens(this.get, this.update);
}

@immutable
class Lens<S, T> extends RawLens<S, S, T, T> {
  const Lens(Function1<S, T> get, Function2<S, T, S> update) : super(get, update);

  Lens<S, Q> then<Q>(Lens<T, Q> lens) =>
      Lens<S, Q>(
            (s) => lens.get(get(s)),
            (s, q) => update(s, lens.update(get(s), q)),
      );

  OLens<S, C> thenO<C>(OLens<T, C> otherLens) {
    return OLens((a) {
      return otherLens.get(get(a));
    }, (a, c) {
      return otherLens.update(get(a), c).map((b) => update(a, b));
    });
  }

  ELens<S, C, E> thenE<C, E>(ELens<T, C, E> otherLens) {
    return ELens((a) {
      return otherLens.get(get(a));
    }, (a, c) {
      return otherLens.update(get(a), c).map((b) => update(a, b));
    });
  }

  Lens<S, Q> thenWithContext<Q>(Lens<T, Q> Function(S context) lensMaker) =>
      Lens<S, Q>(
            (s) => lensMaker(s).get(get(s)),
            (s, q) => update(s, lensMaker(s).update(get(s), q)),
      );

  S map(T Function(T) f, S s) => update(s, f(get(s)));

  FocusedLens<S, T> of(S s) => FocusedLens<S, T>._(s, this);
}

class OLens<A, B> extends RawLens<A, Option<A>, B, Option<B>> {
  const OLens(Function1<A, Option<B>> get, Function2<A, B, Option<A>> set)
      : super(get, set);

  OLens<A, C> oThen<C>(RawLens<B, B, C, C> otherLens) {
    return OLens((a) {
      return get(a).map(otherLens.get);
    }, (a, c) {
      return get(a).bind((b) {
        return update(a, otherLens.update(b, c));
      });
    });
  }

  OLens<A, C> oThenO<C>(OLens<B, C> otherLens) {
    return OLens((a) {
      return get(a).bind(otherLens.get);
    }, (a, c) {
      return get(a).bind((b1) {
        return otherLens.update(b1, c);
      }).bind((b) {
        return update(a, b);
      });
    });
  }
}

class ELens<A, B, E> extends RawLens<A, Either<E, A>, B, Either<E, B>> {
  const ELens(Function1<A, Either<E, B>> getter,
      Function2<A, B, Either<E, A>> setter) : super(getter, setter);

  ELens<A, C, E> eThen<C>(RawLens<B, B, C, C> otherLens) =>
      ELens((a) => get(a).map(otherLens.get), (a, c) =>
          get(a).bind((b) => update(a, otherLens.update(b, c))));

  ELens<A, C, E> eThenE<C>(ELens<B, C, E> otherLens) =>
      ELens((a) => get(a).bind(otherLens.get), (a, c) =>
          get(a).bind((b1) => otherLens.update(b1, c)).bind((b) =>
              update(a, b)));
}

/// A lenses that has been focused on (bound to) a specific instance of the subject.
///
/// Create an instance by using [Lens.of] on a subject.
@immutable
class FocusedLens<S, T> {
  final S _subject;
  final Lens<S, T> _lens;

  const FocusedLens._(this._subject, this._lens);

  FocusedLens<S, Q> then<Q>(Lens<T, Q> lens) =>
      FocusedLens<S, Q>._(_subject, _lens.then(lens));

  FocusedLens<S, Q> thenWithContext<Q>(
      Lens<T, Q> Function(S context) lensMaker) =>
      FocusedLens<S, Q>._(_subject, _lens.thenWithContext(lensMaker));

  T get value => _lens.get(_subject);

  S update(T t) => _lens.update(_subject, t);

  S map(T Function(T) f) => update(f(value));
}

class List$ {
  const List$._();

  static Lens<List<T>, T> atIndex<T>(int i) =>
      Lens<List<T>, T>(
            (s) => s[i],
            (s, t) {
          assert(i >= 0 && i < s.length);
          final newS = List<T>.from(s);
          newS.replaceRange(i, i + 1, [t]);
          return newS;
        },
      );

  static Lens<List<T>, T> first<T>() => atIndex(0);

  static Lens<List<T>, T> where<T>(bool Function(T) predicate) =>
      Lens<List<T>, T>(
            (s) => s.firstWhere(predicate),
            (s, t) {
          final index = s.indexWhere(predicate);
          final newS = List<T>.from(s);
          newS.replaceRange(index, index + 1, [t]);
          return newS;
        },
      );

  static Lens<List<T>, Option<T>> whereOptional<T>(
      bool Function(T) predicate) =>
      Lens<List<T>, Option<T>>(
            (s) => optionOf(s.firstWhere(predicate, orElse: () => null)),
            (s, t) {
          return t.fold(
                () => s, (val) {
            final index = s.indexWhere(predicate);
            if (index < 0) return s;
            final newS = List<T>.from(s);
            newS.replaceRange(index, index + 1, [val]);
            return newS;
          },
          );
        },
      );
}
